
__module_name__ = "_create_label_mask.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu"])


# import packages: -------------------------------------------------------
import licorice_font
import pandas as pd
from anndata import AnnData


# supporting functions: --------------------------------------------------
def _print_label_mask_report(y_mask):
    msg = licorice_font.font_format("Label mask created", ["BOLD"])
    print("{}:\n\n{}".format(msg, y_mask.value_counts()))


# main function: ---------------------------------------------------------
def create_label_mask(
    adata: AnnData, col: str, labels: list, other: str = "other", silent: bool = False
) -> pd.Series:

    """

    Parameters:
    -----------
    adata
        type: anndata.AnnData

    col
        type: str

    labels
        type: list of str

    other
        default: "other"

    silent
        type: bool
        default: False

    Returns:
    --------
    y_mask
        type: pandas.core.series.Series
    """

    # isolate col
    y = adata.obs[col]

    # add the new "other" category
    y_ = y.cat.add_categories(other)

    # rename categories outside of labels as others
    y_ = y_.mask(~y_.isin(labels), other=other)

    # remove the now empty categories that were renamed as others
    y_mask = y_.cat.remove_unused_categories()

    if not silent:
        _print_label_mask_report(y_mask)

    return y_mask



from ._build import build
from ._query import query
from ._save import save
from ._load import load
# specify version: -----------------------------------------------------------------------
__version__ = "0.0.18"





__module_name__ = "_query.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu"])


# import packages: ------------------------------------------------------------
import annoy
import numpy as np


# supporting function: --------------------------------------------------------
def _vector_query(
    annoy_idx: annoy.AnnoyIndex, query_vector: np.ndarray, n_neighbors: int
):
    """
    Parameters:
    -----------
    annoy_idx
        type: Annoy.AnnoyIndex

    query_vector
        type: np.ndarray

    n_neighbors
        type: int

    Returns:
    --------
    nn_idx
        Nearest neighbors of the passed vector from the annoy_idx.
        type: np.ndarray
    """
    return annoy_idx.get_nns_by_vector(query_vector, n_neighbors)


# primary function: -----------------------------------------------------------
def query(
    annoy_idx: annoy.AnnoyIndex, X_query: np.ndarray, n_neighbors: int = 20
) -> np.ndarray:

    """
    Query annoy index by vector.

    Parameters:
    -----------
    annoy_idx
        Spotify.AnnoyIndex of nearest neighbors built on X_ref.
        type: Annoy.AnnoyIndex

    X_query
        Vector of items x n_dim (on which the graph was built). To query
        the neigbor graph.
        type: np.ndarray

    n_neighbors
        Number of neighbors by which to query the graph.
        type: int
        default: 20

    Returns:
    --------
    nn_query
        item indices of the nearest neighbors to X_query (from X_ref)
        type: np.ndarray
    """

    nn_idx_list = []

    query_size = X_query.shape[0]
    for i in range(query_size):
        nn_idx = _vector_query(annoy_idx, X_query[i], n_neighbors)
        nn_idx_list.append(nn_idx)

    return np.array(nn_idx_list)


import anndata
import pandas as pd
import numpy as np
from tqdm.notebook import tqdm
import vinplots
import matplotlib.pyplot as plt


from .._core._SpotifyAnnoy import SpotifyAnnoy as kNN
from ._annotate_adata_bool_idx import add_bool_idx

def _plot_smoothing(adata, score_idx, scores, n_iters):

    X_umap = adata[score_idx].obsm["X_umap"]

    nplots = n_iters + 1
    fig, axes = vinplots.quick_UMAP(nplots=nplots, ncols=nplots)

    c_idx = np.argsort(scores[0])
    axes[0].scatter(
        X_umap[:, 0], X_umap[:, 1], c="lightgrey", s=2, rasterized=True, zorder=0
    )
    axes[0].scatter(
        X_umap[c_idx, 0],
        X_umap[c_idx, 1],
        c=scores[0][c_idx],
        rasterized=True,
        zorder=1,
    )
    axes[0].set_title("Mapped (unsmoothed)")

    for i in range(1, nplots):
        axes[i].scatter(
            X_umap[:, 0],
            X_umap[:, 1],
            c="lightgrey",
            rasterized=True,
            zorder=0,
        )
        c_idx = np.argsort(scores[i])
        axes[i].scatter(
            X_umap[c_idx, 0],
            X_umap[c_idx, 1],
            c=scores[i][c_idx],
            rasterized=True,
            zorder=1,
        )
        axes[i].set_title("iter: {}".format(i))
    plt.show()


def smooth(
    adata: anndata.AnnData,
    score: pd.Series,
    plot=False,
    n_iters: int = 5,
    n_neighbors: int = 20,
    func: bool = False,
) -> np.ndarray:

    idx, score = score.index.astype(int), score.values.astype(float)
    curr = np.zeros(score.shape)
    scores, prev = [score], score

    add_bool_idx(adata, idx, key_added="tmp_bool", silent=True)

    kNN_Graph = kNN(adata)
    kNN_Graph.build(subset="tmp_bool")

    X_query = adata[idx].obsm["X_pca"]

    for _ in tqdm(range(n_iters)):
        X_nn = kNN_Graph.query(X_query)
        curr = prev[X_nn]
        if func:
            curr = func(curr)
        else:
            curr = curr.mean(1)
        prev = curr
        scores.append(curr)

    if plot:
        _plot_smoothing(adata, idx, scores, n_iters)

    return curr



__module_name__ = "__init__.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


# VERSION: --------------------------------------------------------------------
__version__ = "0.0.18"


# Core module and functions: --------------------------------------------------
from ._core._SpotifyAnnoy import SpotifyAnnoy as kNN
from ._core._idx_funcs import *


# Data formatting and other interactive modules: ------------------------------
from ._interactive_funcs import *


# Data mapping functions: -----------------------------------------------------
from ._map._kNN_map_source_to_target import map_source_to_target
from ._map._iterative_kNN_smoothing import smooth



__module_name__ = "_save.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu"])


# import packages: ------------------------------------------------------------
import licorice_font
import annoy
import pydk
import os


# supporting functions: -------------------------------------------------------
def _save_message(ann_fname):

    note = licorice_font.font_format("NOTE", ["BLUE"])
    msg = licorice_font.font_format("Saving annoy idx to", ["BOLD"])
    print(" - [{}] | {}:\n\n\t{}".format(note, msg, ann_fname))


def _mk_idx_fname(dim, n_neighbors, n_trees):
    msg = "annoyance.annoy_idx.{}_dim.{}_neighbors.{}_trees"
    return msg.format(dim, n_neighbors, n_trees)


# main function: --------------------------------------------------------------
def save(
    idx: annoy.AnnoyIndex,
    n_dim: int,
    n_neighbors: int,
    save_dir: str = "./",
    silent=False,
) -> None:
    
    """
    Save AnnoyIndex to /path/to/save_dir/annoyance/AnnoyIndex.ann

    Parameters:
    -----------
    annoy_idx
        Spotify.AnnoyIndex of nearest neighbors built on X_ref.
        type: Annoy.AnnoyIndex
    
    n_dim
        type: int
    
    n_neighbors
        Number of neighbors by which to query the graph.
        type: int
        default: 20

    save_dir
        Parant save directory path
        type: str
        default: "./"

    silent
        Silences update messages.
        type: bool
        default: False

    Returns:
    --------
    None, saves .ann file to disk.
    """

    annoy_save_dir = os.path.join(save_dir, "annoyance")
    pydk.mkdir_flex(annoy_save_dir)

    base_fname = _mk_idx_fname(n_dim, n_neighbors, idx.get_n_trees())
    annoy_idx_base_path = os.path.join(annoy_save_dir, base_fname)
    ann_fname = "{}.ann".format(annoy_idx_base_path)
    idx.save(ann_fname)
    if not silent:
        _save_message(ann_fname)


import anndata
import pandas as pd
import numpy as np
import time


def _count_values(col: pd.Series) -> dict:
    return col.value_counts().to_dict()

def _max_count(col: pd.Series) -> str:
    return col.value_counts().idxmax()

def count_obs_neighbors(
    adata: anndata.AnnData, query_result: np.ndarray, obs_key: str, max_only=False,
) -> (list([dict, dict, ...]), float):

    """
    Queried kNN returns index of nearest neighbors. Query the adata.obs
    table using this index and count returned values (of a column).
    
    Parameters:
    -----------
    adata
        type: anndata.AnnData

    query_result
        result of SpotifyAnnoy.query()
        type: np.ndarray

    obs_key
        type: str
        
    Returns:
    --------
    value_counts
        Value counts for each set of neighbors for a given input.
    
    query_time
        Time taken to perform the query.
        type: float
    
    
    Notes:
    ------
    (1) This implementation is much faster than a loop. Querying AnnData
        is very slow when done many times;best to do it only once.
    """

    query_time_start = time.time()

    nn_adata = adata[query_result.flatten()]
    # nn_idx = np.repeat(range(len(query_result)), 20)
    query_df = pd.DataFrame(
        nn_adata.obs[obs_key].to_numpy().reshape(query_result.shape).T
    )
    
    if max_only:
        value_counts = [
            _max_count(query_df[i]) for i in query_df.columns
        ] # list of values
    else:
        value_counts = [
            _count_values(query_df[i]) for i in query_df.columns
        ]  # list of dicts

    query_time = time.time() - query_time_start

    del nn_adata

    return value_counts, query_time



__module_name__ = "_build.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu"])


# import packages: ------------------------------------------------------------
import annoy


def build(
    x, metric: str = "euclidean", n_trees: int = 10
) -> annoy.AnnoyIndex:
    
    """
    Build a Spotify.annoy kNN index
    
    Parameters:
    -----------
    x
        input matrix
        type: numpy.ndarray
    
    metric
        distance metric
        type: str
        default: "euclidean"
    
    n_trees
        number of trees.
        type: int
        default: 10
        
    Returns:
    --------
    annoy_idx
        type: annoy.AnnoyIndex
    
    Notes:
    ------
    (1) Loops through each item. Appends n_dim feature vector to AnnoyIndex.
    """    

    annoy_idx = annoy.AnnoyIndex(x.shape[1], metric)

    for i in range(x.shape[0]):
        annoy_idx.add_item(i, x[i])

    annoy_idx.build(n_trees)

    return annoy_idx




__module_name__ = "_SpotifyAnnoy.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu"])


# import packages: ------------------------------------------------------------
import os, inspect
import numpy as np
import annoy
import anndata


# import local dependencies: --------------------------------------------------
from . import idx
from annoyance._interactive_funcs import *


# Main module class: ----------------------------------------------------------
class SpotifyAnnoy:
    def __init__(
        self,
        adata: anndata.AnnData,
        use_key: str = "X_pca",
        obs_key: str = None,
        n_neighbors: int = 20,
        n_trees: int = 10,
        metric: str = "euclidean",
    ):
        """
        Initialize the interactive module for building kNN graphs with AnnData
        using Spotify's super-fast Annoy Library.
        
        adata
            AnnData object for single-cell data
            type: anndata.AnnData
        
        use_key
            Key that specififes which matrix of AnnData object should be used
            for input items.
            Must be "X" or contained in adata.obsm_keys()
            type: str

        obs_key
            type: str
        
        n_neighbors [ optional, if defined above ]
            Number of neighbors by which to query the graph
            type: int
            default: None 
        
        n_trees
            number of trees 
            type: int
            default: 10
            
        metric
            distance metric; options: "angular", "euclidean", "manhattan", "hamming", or "dot".
            type: str
            default: "euclidean"
        """

        self._register_params(locals())
        self.X_ref = use_X(self._adata, use_key=self._use_key)

    def _register_params(self, local_params):

        sig = inspect.signature(SpotifyAnnoy)
        for key, value in sig.parameters.items():
            self.__setattr__("_{}".format(key), local_params[value.name])

    def build(
        self,
        subset: str = None,
        X_ref: np.ndarray = None,
        metric: str = "euclidean",
        n_trees: int = 10,
    )->annoy.AnnoyIndex:
        
        """
        Build a Spotify.annoy kNN index

        Parameters:
        -----------
        subset
            Graph will be built on the subset of data wherein: adata[adata.obs[col] == True]
            type: str
            default: None
            
        X_ref [ optional if defined during __init__ ]
            input matrix
            type: numpy.ndarray

        metric [ optional if defined during __init__ ]
            distance metric; options: "angular", "euclidean", "manhattan", "hamming", or "dot".
            type: str
            default: "euclidean"

        n_trees [ optional if defined during __init__ ]
            number of trees 
            type: int
            default: 10

        Returns:
        --------
        annoy_idx
            type: annoy.AnnoyIndex

        Notes:
        ------
        (1) Loops through each item. Appends n_dim feature vector to AnnoyIndex.
        """

        # update values if necessary --------------------------------------------------
        if X_ref:
            self.X_ref = X_ref
            
        elif subset:
            adata_subset = self._adata[self._adata.obs[subset]]
            self.X_ref = use_X(adata_subset, use_key=self._use_key)
            
        if metric != self._metric:
            self._metric = metric
            
        if n_trees != self._n_trees:
            self._n_trees = n_trees        

        self.idx = idx.build(self.X_ref, metric=self._metric, n_trees=self._n_trees)

    def query(
        self, X_query: np.ndarray, n_neighbors = None, return_query: bool = True
    ) -> np.ndarray:
        
        
        """
        Query annoy index by vector - typically of unseen data.

        Parameters:
        -----------
        X_query
            Vector of items x n_dim (on which the graph was built). To query
            the neigbor graph.
            type: np.ndarray

        n_neighbors [ optional, if defined above ]
            Number of neighbors by which to query the graph
            type: int
            default: None 
            
        return_query
            Return the np.ndarray outside of the class
            type: bool
            default: True

        Returns:
        --------
        nn_query
            item indices of the nearest neighbors to X_query (from X_ref)
            type: np.ndarray
        """
        
        if not n_neighbors:
            n_neighbors = self._n_neighbors            
        
        self._nn_query = idx.query(self.idx, X_query, n_neighbors)
        if return_query:
            return self._nn_query

    def count(self, obs_key: str, max_only: bool = False) -> (list([dict, dict, ...]), float):
        
        """
        Queried kNN returns index of nearest neighbors. Query the adata.obs
        table using this index and count returned values (of a column).

        Parameters:
        -----------
        obs_key
            type: str

        Pre-defined arguments:
        ----------------------
        adata
            AnnData objects. Calls self._adata
            type: anndata.AnnData

        query_result
            result of SpotifyAnnoy.query(). Calls self._nn_query
            type: np.ndarray
        
        Returns:
        --------
        value_counts
            Value counts for each set of neighbors for a given input.
            Updated as: self._query_value_counts

        query_time
            Time taken to perform the query.
            type: float
            Updated as self._count_time


        Notes:
        ------
        (1) This implementation is much faster than a loop. Querying AnnData
            is very slow when done many times;best to do it only once.
        """
        
        self._query_value_counts, self._count_time = count_obs_neighbors(
            self._adata, self._nn_query, obs_key, max_only,
        )
        
    def load(
        self, path: str, n_features: int = None, metric: str = "euclidean"
    ) -> annoy.AnnoyIndex:

        """
        Load a previously saved AnnoyIndex.

        Parameters:
        -----------
        path
            file path  to annoy idx (.ann)

        n_features
            type: int

        metric
            Options: "angular", "euclidean", "manhattan", "hamming", or "dot".
            type: str
            default: "euclidean"

        Returns:
        --------
        annoy_idx
            Updates existing `self.idx` with saved kNN graph.
            type: annoy.AnnoyIndex

        Notes:
        ------
        (1) `self._n_items` which is also return as a method of the annoy_idx corresponds to the number 
        of cells on which classifier was trained.
        """

        if not n_features:
            n_features = self.X_ref.shape[1]

        self.idx = idx.load(path, n_features, metric)

    def save(self, save_dir: str = "./", silent: bool = False)->None:
        
        """
        Save AnnoyIndex to /path/to/save_dir/annoyance/AnnoyIndex.ann

        Parameters:
        -----------
        save_dir
            Parant save directory path
            type: str
            default: "./"

        silent
            Silences update messages.
            type: bool
            default: False

        Returns:
        --------
        None, saves .ann file to disk.
        """
        
        n_dim = self.X_ref.shape[1]
        idx.save(self.idx, n_dim, self._n_neighbors, save_dir, silent)



import licorice_font
import numpy as np

def _print_key_added(key_added):

    note = licorice_font.font_format("NOTE", ["BLUE"])
    key_print = licorice_font.font_format(key_added, ["BOLD"])
    print(" - [{}] | obs_key added: {}".format(note, key_print))
    
    
    
def add_bool_idx(adata, idx, key_added="bool_idx", silent=False):
    
#     _annotate_adata_bool_idx

    tmp = np.zeros(len(adata))
    tmp[idx] = 1
    adata.obs[key_added] = tmp.astype(bool)

    if not silent:
        _print_key_added(key_added)




__module_name__ = "__init__.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])



# specify version: -----------------------------------------------------------------------
__version__ = "0.0.18"

# -----------------------------------------------------------------------------
from ._binominal_test_train_assignment import binominal_test_train_assignment
from ._count_obs_neighbors import count_obs_neighbors
from ._create_label_mask import create_label_mask
from ._use_X import use_X



__module_name__ = "__init__.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])



# specify version: -----------------------------------------------------------------------
__version__ = "0.0.18"

# Modules and functions: ------------------------------------------------------
from ._kNN_map_source_to_target import map_source_to_target
from ._iterative_kNN_smoothing import smooth
from ._annotate_adata_bool_idx import add_bool_idx



__module_name__ = "_load.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu"])


# import packages: ------------------------------------------------------------
import annoy


# primary function: -----------------------------------------------------------
def load(
    path: str, n_features: int, metric: str = "euclidean"
) -> annoy.AnnoyIndex:

    """
    Load a previously saved AnnoyIndex.

    Parameters:
    -----------
    path
        file path  to annoy idx (.ann)

    n_features
        type: int

    metric
        type: str
        default: "euclidean"

    Returns:
    --------
    annoy_idx
        kNN graph
        type: annoy.AnnoyIndex

    Notes:
    ------
    """

    annoy_idx = annoy.AnnoyIndex(n_features, metric)
    annoy_idx.load(path)

    return annoy_idx



__module_name__ = "_binominal_test_train_assignment.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu"])


# import packages: ------------------------------------------------------------
import anndata
import numpy as np


# primary function: -----------------------------------------------------------
def binominal_test_train_assignment(adata: anndata.AnnData, p: float = 0.5) -> None:

    """
    Binomially assign test/train split.

    Parameters:
    -----------
    adata
        AnnData object
        type: anndata.AnnData

    p
        probability of training-set selection.
        type: float
        default: 0.5

    Returns:
    --------
    None
        adata is modified in-place, updated with adata.obs['kNN_test'] and adata.obs['kNN_train']
    """

    obs_df = adata.obs.copy()

    train = np.random.binomial(1, p, len(obs_df)).astype(bool)
    test = np.invert(train)

    obs_df["kNN_train"] = train
    obs_df["kNN_test"]  = test

    adata.obs = obs_df



__module_name__ = "_use_X.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu"])


# import packages: -------------------------------------------------------
import licorice_font
import numpy


# supporting functions: --------------------------------------------------
def _is_numpy_array(x):
    return x.__class__ is numpy.ndarray


def _as_numpy_array(x):
    if _is_numpy_array(x):
        return x
    else:
        return x.toarray()

def _print_help_message(adata, use_key):
    msg = " - [{}] | {} is not a suitable array key for this dataset. {}:\n".format(
        licorice_font.font_format("NOTE", ["BLUE"]),
        licorice_font.font_format(use_key, ["BOLD", "RED"]),
        licorice_font.font_format("Choose from", ["BOLD"]),
    )
    print(msg)
    available = [licorice_font.font_format(key, ["BOLD"]) for key in adata.obsm_keys()]
    licorice_font.underline("Counts matrix:", ["BOLD", "GREEN"], n_newline=0)
    print(" - {}\n".format(licorice_font.font_format("X", ["BOLD"])))
    licorice_font.underline("adata.obsm_keys:", ["BOLD", "GREEN"], n_newline=0)
    for obsm_key in available:
        print(" - {}".format(obsm_key))
        
# main function: fetch data (X_ref): -------------------------------------
def use_X(adata, use_key="X"):

    """
    Return data from AnnData as numpy array (if not already).

    Parameters:
    -----------
    adata
        AnnData object
        type: anndata._core.anndata.AnnData

    use_key
        "X" or key in adata.obsm_keys(), adata.obs_keys(),
        type: str
        default: "X"

    Returns:
    --------
    X
        Formatted data matrix.
        type: np.ndarray
    """

    if use_key == "X":
        return _as_numpy_array(adata.X)

    elif use_key in adata.obsm_keys():
        return _as_numpy_array(adata.obsm[use_key])

    else:
        _print_help_message(adata, use_key)




__module_name__ = "_kNN_map_source_to_target.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


# import packages: ------------------------------------------------------------
import pandas as pd
import numpy as np
import anndata


# import local dependencies: --------------------------------------------------
from .._core._SpotifyAnnoy import SpotifyAnnoy as kNN


# Supporting functions: -------------------------------------------------------
def _add_source_target_exclude_col(
    adata,
    source_idx,
    target_idx,
):

    tmp = np.full(len(adata), "exclude")
    tmp[source_idx] = "source"
    tmp[target_idx] = "target"

    adata.obs["kNN_mapping"] = pd.Categorical(tmp)

    adata.obs["source"] = adata.obs["kNN_mapping"] == "source"
    adata.obs["target"] = adata.obs["kNN_mapping"] == "target"

    adata.obs.drop("kNN_mapping", axis=1, inplace=True)


def _source_target_query(adata, use_key):

    """build and query kNN index"""

    G = kNN(adata)
    G.build(subset="source")
    X_query = adata[adata.obs["target"]].obsm[use_key].toarray()

    return G.query(X_query)


def _merge_ref_map_values(adata, obs_key, X_nn):

    ref_values = adata[adata.obs["source"]].obs[obs_key].values
    map_values = ref_values[X_nn].mean(1)

    tmp = np.full(len(adata), None)
    tmp[adata.obs["source"]] = ref_values
    tmp[adata.obs["target"]] = map_values
    adata.obs["{}_mapped".format(obs_key)] = tmp

    adata.obs.drop(["source", "target"], axis=1, inplace=True)


# Main module function: -------------------------------------------------------
def map_source_to_target(
    adata: anndata.AnnData,
    source_idx: pd.Index,
    target_idx: pd.Index,
    obs_key: str,
    use_key="X_pca",
):

    """
    Using a kNN graph, map a value in a source set of cells to a target set of cells.

    Parameters:
    -----------
    adata
        type: anndata.AnnData

    source_idx [ required ]
        type: pandas.Index

    target_idx [ required ]
        type: pandas.Index

    obs_key [ required ]
        type: str

    use_key
        type: str
        default: "X_pca"

    Returns:
    --------
    None, updates adata.obs with "{obs_key}_mapped"

    Notes:
    ------
    (1) annotates source/target in the adata.obs table
    (2) build (based on the source) a kNN graph and query (using the target)
    (3) merge mapped / references values
    """
    _add_source_target_exclude_col(adata, source_idx, target_idx)
    X_nn = _source_target_query(adata, use_key)
    _merge_ref_map_values(adata, obs_key, X_nn)
    
